#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import decimal

import mock
import pytest

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.signals as signals

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order, sell_limit_order, stop_loss_limit_order, stop_loss_buy_order

import octobot_trading.personal_data as personal_data


@pytest.fixture
def trading_signal_bundle_builder():
    builder = signals.TradingSignalBundleBuilder(
        "identifier",
        "strategy_name",
    )
    builder.logger = mock.Mock(debug=mock.Mock(), error=mock.Mock())
    return builder


def test_build(trading_signal_bundle_builder):
    with mock.patch.object(trading_signal_bundle_builder, "_pack_referenced_orders_together", mock.Mock()) \
         as _pack_referenced_orders_together_mock:
        trading_signal_bundle = trading_signal_bundle_builder.build()
        assert trading_signal_bundle.identifier == "identifier"
        assert trading_signal_bundle.signals == []
        _pack_referenced_orders_together_mock.assert_called_once()


def test_sort(trading_signal_bundle_builder, buy_limit_order):
    assert trading_signal_bundle_builder.signals == []
    assert trading_signal_bundle_builder.sort_signals().signals == []
    order_ids = [
        f"order_id_{i}"
        for i in range(5)
    ]
    buy_limit_order.order_id = order_ids[0]
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    # add new order (orders are based on order_id)
    buy_limit_order.order_id = order_ids[1]
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, buy_limit_order.exchange_manager)
    buy_limit_order.order_id = order_ids[2]
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, buy_limit_order.exchange_manager)
    buy_limit_order.order_id = order_ids[3]
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="2%")
    buy_limit_order.order_id = order_ids[4]
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="3%")

    origin_signals = copy.copy(trading_signal_bundle_builder.signals)
    assert order_ids == [signal.content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] for signal in origin_signals]
    builder = trading_signal_bundle_builder.sort_signals()
    assert builder is trading_signal_bundle_builder
    sorted_signals = builder.signals
    sorted_ids = [order_ids[1], order_ids[2], order_ids[0], order_ids[3], order_ids[4]]
    assert sorted_ids == [signal.content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] for signal in sorted_signals]


def test_add_created_order(trading_signal_bundle_builder, buy_limit_order):
    with pytest.raises(errors.InvalidArgumentError):
        trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager)
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalCommonsAttrs.ACTION.value] \
           is enums.TradingSignalOrdersActions.CREATE.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == \
           buy_limit_order.order_id
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] == "1%"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None

    # add the same order: do not add it twice
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    assert len(trading_signal_bundle_builder.signals) == 1

    # update the same order
    buy_limit_order.order_type = enums.TraderOrderType.SELL_LIMIT
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_position="2%")
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.SELL_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "2%"

    # add new order (orders are based on order_id)
    previous_order_id = buy_limit_order.order_id
    buy_limit_order.order_id = "other_id"
    buy_limit_order.order_type = enums.TraderOrderType.STOP_LOSS_LIMIT
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_position="50")
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == \
           previous_order_id
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.SELL_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "2%"
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == "other_id"
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "50"


def test_add_order_to_group(trading_signal_bundle_builder, buy_limit_order):
    # no order_group
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, buy_limit_order.exchange_manager)
    # ensure properly added
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalCommonsAttrs.ACTION.value] \
           is enums.TradingSignalOrdersActions.ADD_TO_GROUP.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == \
           buy_limit_order.order_id
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None

    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] is None

    order_group = personal_data.OneCancelsTheOtherOrderGroup(
        "group_name",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(order_group)
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.OneCancelsTheOtherOrderGroup.__name__

    # add the same order: do not add it twice
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] == order_group.name
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == order_group.__class__.__name__

    # update the same order
    other_order_group = personal_data.BalancedTakeProfitAndStopOrderGroup(
        "group_name_2",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(other_order_group)
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name_2"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.BalancedTakeProfitAndStopOrderGroup.__name__

    # add new order (orders are based on order_id)
    buy_limit_order.order_id = "other_id"
    buy_limit_order.add_to_order_group(order_group)
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name_2"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.BalancedTakeProfitAndStopOrderGroup.__name__
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name"
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.OneCancelsTheOtherOrderGroup.__name__


def test_add_edited_order(trading_signal_bundle_builder, buy_limit_order):
    # no updated argument
    with pytest.raises(errors.InvalidArgumentError):
        trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager)

    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_target_amount="1%")
    # ensure properly added
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalCommonsAttrs.ACTION.value] \
           is enums.TradingSignalOrdersActions.EDIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == \
           buy_limit_order.order_id
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] == "1%"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] == "1%"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0

    # add the same order: do not add it twice, update existing order
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_target_position="1%")
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "1%"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == "1%"
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_limit_price=decimal.Decimal("1"))
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 1.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_stop_price=decimal.Decimal("1"))
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 1.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_current_price=decimal.Decimal("1"))
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 1.0

    # add new order (orders are based on order_id)
    buy_limit_order.order_id = "other_id"
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, buy_limit_order.exchange_manager, updated_target_position="1%a")
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 1.0
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == None
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == "1%a"


@pytest.mark.asyncio
async def test_add_cancelled_order(trading_signal_bundle_builder, buy_limit_order):
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalCommonsAttrs.ACTION.value] \
           is enums.TradingSignalOrdersActions.CANCEL.value
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ORDER_ID.value] == \
           buy_limit_order.order_id

    # add the same order: do not add it twice
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, buy_limit_order.exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 1

    # add new order (orders are based on order_id)
    # cancel order first to be sure it can be added
    exchange_manager = buy_limit_order.exchange_manager
    await buy_limit_order.exchange_manager.trader.cancel_order(buy_limit_order)
    buy_limit_order.clear()
    buy_limit_order.order_id = "other_id"
    buy_limit_order.order_type = enums.TraderOrderType.STOP_LOSS_LIMIT
    trading_signal_bundle_builder.add_created_order(buy_limit_order, exchange_manager, target_position="50")
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value

    # add an orders via create action: it gets popped out of the orders list as there is no point creating
    # it and cancelling it right away
    buy_limit_order.order_id = "buy_other_id"
    buy_limit_order.order_type = enums.TraderOrderType.BUY_MARKET
    trading_signal_bundle_builder.add_created_order(buy_limit_order, exchange_manager, target_amount="1")
    assert len(trading_signal_bundle_builder.signals) == 3
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value

    # BUY_MARKET order got removed from orders
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value

    # adding BUY_MARKET as add to group order
    other_order_group = personal_data.BalancedTakeProfitAndStopOrderGroup(
        "group_name_2",
        exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(other_order_group)
    trading_signal_bundle_builder.add_order_to_group(buy_limit_order, exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 3
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.ADD_TO_GROUP.value

    # BUY_MARKET order sill in order but action is not cancel from orders
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 3
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value

    # adding BUY_MARKET as edited order
    buy_limit_order.order_id = "edit_buy_other_id"
    trading_signal_bundle_builder.add_edited_order(buy_limit_order, exchange_manager, updated_target_amount="1")
    assert len(trading_signal_bundle_builder.signals) == 4
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value
    assert trading_signal_bundle_builder.signals[3].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[3].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.EDIT.value

    # BUY_MARKET order sill in order but action is not cancel from orders
    trading_signal_bundle_builder.add_cancelled_order(buy_limit_order, exchange_manager)
    assert len(trading_signal_bundle_builder.signals) == 4
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[2].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value
    assert trading_signal_bundle_builder.signals[3].content[enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert trading_signal_bundle_builder.signals[3].content[enums.TradingSignalCommonsAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value


def test_pack_referenced_orders_together(trading_signal_bundle_builder,
                                         buy_limit_order, sell_limit_order, stop_loss_limit_order, stop_loss_buy_order):
    buy_limit_order.order_id = "0"
    sell_limit_order.order_id = "1"
    stop_loss_limit_order.order_id = "2"
    trading_signal_bundle_builder.add_created_order(buy_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    trading_signal_bundle_builder.add_created_order(sell_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    trading_signal_bundle_builder.add_created_order(stop_loss_limit_order, buy_limit_order.exchange_manager, target_amount="1%")
    assert len(trading_signal_bundle_builder.signals) == 3
    assert all(signal.content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] == []
               for signal in trading_signal_bundle_builder.signals)
    pre_pack_signals = copy.copy(trading_signal_bundle_builder.signals)
    trading_signal_bundle_builder._pack_referenced_orders_together()
    # no order to be packed, no change
    assert len(trading_signal_bundle_builder.signals) == 3
    assert all(signal.content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] == []
               for signal in trading_signal_bundle_builder.signals)

    # missing id
    trading_signal_bundle_builder.logger.debug.assert_not_called()
    pre_pack_signals[2].content[enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value] = "11"
    trading_signal_bundle_builder._pack_referenced_orders_together()
    trading_signal_bundle_builder.logger.debug.assert_called_once()
    assert "triggering order not found" in trading_signal_bundle_builder.logger.debug.call_args[0][0]
    trading_signal_bundle_builder.logger.debug.reset_mock()
    # no signals update
    assert trading_signal_bundle_builder.signals == pre_pack_signals

    # bundle stop loss with buy limit
    pre_pack_signals[2].content[enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value] = "0"
    trading_signal_bundle_builder._pack_referenced_orders_together()
    # no order to be packed, no change
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][0] \
           is pre_pack_signals[2].content
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] \
           == []
    trading_signal_bundle_builder.logger.debug.assert_not_called()

    # reset signals
    trading_signal_bundle_builder.signals = pre_pack_signals
    for signal in trading_signal_bundle_builder.signals:
        signal.content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] = []
    # also chain sell limit to buy limit
    pre_pack_signals[1].content[enums.TradingSignalOrdersAttrs.CHAINED_TO.value] = "0"
    trading_signal_bundle_builder._pack_referenced_orders_together()
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][0] \
           is pre_pack_signals[1].content
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][1] \
           is pre_pack_signals[2].content
    trading_signal_bundle_builder.logger.debug.assert_not_called()

    # reset signals
    trading_signal_bundle_builder.signals = pre_pack_signals
    for signal in trading_signal_bundle_builder.signals:
        signal.content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] = []
    # also chain sell limit to buy limit
    pre_pack_signals[1].content[enums.TradingSignalOrdersAttrs.CHAINED_TO.value] = "0"
    trading_signal_bundle_builder._pack_referenced_orders_together()
    assert len(trading_signal_bundle_builder.signals) == 1
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][0] \
           is pre_pack_signals[1].content
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][1] \
           is pre_pack_signals[2].content
    trading_signal_bundle_builder.logger.debug.assert_not_called()

    # reset signals
    trading_signal_bundle_builder.signals = pre_pack_signals
    for signal in trading_signal_bundle_builder.signals:
        signal.content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value] = []
    # add stop_loss_buy_order in the same group as stop_loss_sell_order
    trading_signal_bundle_builder.add_created_order(stop_loss_buy_order, buy_limit_order.exchange_manager, target_amount="1%")
    trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value] = None
    trading_signal_bundle_builder.signals[2].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] = "grp"
    trading_signal_bundle_builder.signals[3].content[enums.TradingSignalOrdersAttrs.GROUP_ID.value] = "grp"
    pre_pack_signals = copy.copy(trading_signal_bundle_builder.signals)
    trading_signal_bundle_builder._pack_referenced_orders_together()
    trading_signal_bundle_builder.logger.debug.assert_not_called()
    assert len(trading_signal_bundle_builder.signals) == 2
    assert trading_signal_bundle_builder.signals[0].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][0] \
           is pre_pack_signals[1].content
    assert trading_signal_bundle_builder.signals[1].content[enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value][0] \
           is pre_pack_signals[3].content
    
    trading_signal_bundle_builder.logger.error.assert_not_called()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

import octobot_commons.logging as logging
import octobot_commons.symbols as symbol_util

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data.portfolios.value_converter as value_converter


class PortfolioValueHolder:
    """
    PortfolioValueHolder calculates the current and the origin portfolio value in reference market for each updates
    """

    def __init__(self, portfolio_manager):
        self.portfolio_manager = portfolio_manager
        self.logger = logging.get_logger(f"{self.__class__.__name__}"
                                         f"[{self.portfolio_manager.exchange_manager.exchange_name}]")
        self.value_converter = value_converter.ValueConverter(self.portfolio_manager)

        self.portfolio_origin_value = constants.ZERO
        self.portfolio_current_value = constants.ZERO

        # values in decimal.Decimal
        self.origin_portfolio = None

        # values in decimal.Decimal
        self.origin_crypto_currencies_values = {}
        self.current_crypto_currencies_values = {}

    def reset_portfolio_values(self):
        self.portfolio_origin_value = constants.ZERO
        self.portfolio_current_value = constants.ZERO

        self.origin_portfolio = None

        self.origin_crypto_currencies_values = {}
        self.current_crypto_currencies_values = {}

    def update_origin_crypto_currencies_values(self, symbol, mark_price):
        """
        Update origin cryptocurrencies value
        :param symbol: the symbol to update
        :param mark_price: the symbol mark price value in decimal.Decimal
        :return: True if the origin portfolio should be recomputed
        """
        currency, market = symbol_util.parse_symbol(symbol).base_and_quote()
        # update origin values if this price has relevant data regarding
        # the origin portfolio (using both quote and base)
        origin_crypto_currencies_with_values = set(self.origin_crypto_currencies_values.keys())
        origin_currencies_should_be_updated = (
            (
                currency not in origin_crypto_currencies_with_values
                and currency != self.portfolio_manager.reference_market
            )
            or
            (
                market not in origin_crypto_currencies_with_values
                and market != self.portfolio_manager.reference_market
            )
        )
        self.value_converter.update_last_price(symbol, mark_price)
        if origin_currencies_should_be_updated:
            # Will fail if symbol doesn't have a price in
            # self.origin_crypto_currencies_values and therefore
            # requires the origin portfolio value to be recomputed 
            # using this price info in case this price is relevant
            if market == self.portfolio_manager.reference_market:
                self.origin_crypto_currencies_values[currency] = mark_price
            elif currency == self.portfolio_manager.reference_market:
                self.origin_crypto_currencies_values[market] = constants.ONE / mark_price
            else:
                try:
                    converted_value = self.value_converter.try_convert_currency_value_using_multiple_pairs(
                        currency, self.portfolio_manager.reference_market, constants.ONE, []
                    )
                    if converted_value is not None:
                        self.origin_crypto_currencies_values[currency] = converted_value
                except (errors.MissingPriceDataError, errors.PendingPriceDataError):
                    pass
        return origin_currencies_should_be_updated

    def get_current_crypto_currencies_values(self):
        """
        Return the current crypto-currencies values
        :return: the current crypto-currencies values
        """
        if not self.current_crypto_currencies_values:
            self._update_portfolio_and_currencies_current_value()
        return self.current_crypto_currencies_values

    def get_current_holdings_values(self):
        """
        Get holdings ratio for each currency
        :return: the holdings ratio dictionary
        """
        holdings = self.get_current_crypto_currencies_values()
        return {
            currency: self._get_currency_value(self.portfolio_manager.portfolio.portfolio, currency, holdings)
            for currency in holdings.keys()
        }

    def get_assets_holdings_value(self, assets, target_unit, init_price_fetchers=False):
        total_value = constants.ZERO
        for asset, asset_holdings in self.portfolio_manager.portfolio.portfolio.items():
            if asset not in assets:
                continue
            try:
                total_value += self.value_converter.evaluate_value(
                    asset, asset_holdings.total, raise_error=True,
                    target_currency=target_unit, init_price_fetchers=init_price_fetchers
                )
            except errors.MissingPriceDataError:
                self.logger.info(f"Missing {asset} price conversion, ignoring {float(asset_holdings.total)} holdings.")
        return total_value

    def get_traded_assets_holdings_value(self, unit):
        assets = set()
        for symbol in self.portfolio_manager.exchange_manager.exchange_config.traded_symbols:
            assets.add(symbol.base)
            assets.add(symbol.quote)
        return self.get_assets_holdings_value(
            assets, unit
        )

    def get_holdings_ratio(self, currency, traded_symbols_only=False, include_assets_in_open_orders=False):
        if traded_symbols_only:
            total_holdings_value = self.get_traded_assets_holdings_value(self.portfolio_manager.reference_market)
        else:
            # consider all assets for total_holdings_value
            total_holdings_value = self.portfolio_current_value
        if not total_holdings_value:
            return constants.ZERO
        currency_holdings = self.portfolio_manager.portfolio.get_currency_portfolio(currency).total
        if include_assets_in_open_orders:
            # add assets in open orders to currency_holdings
            assets_in_open_orders = constants.ZERO
            for order in self.portfolio_manager.exchange_manager.exchange_personal_data.orders_manager.get_open_orders():
                symbol = symbol_util.parse_symbol(order.symbol)
                if order.side is enums.TradeOrderSide.BUY and symbol.base == currency:
                    assets_in_open_orders += order.origin_quantity
                elif order.side is enums.TradeOrderSide.SELL and symbol.quote == currency:
                    assets_in_open_orders += order.total_cost
            currency_holdings += assets_in_open_orders
        # compute ratio
        current_holdings_value = self.value_converter.evaluate_value(currency, currency_holdings)
        return current_holdings_value / total_holdings_value

    def handle_profitability_recalculation(self, force_recompute_origin_portfolio):
        """
        Initialize values required by portfolio profitability to perform its profitability calculation
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        self._update_portfolio_and_currencies_current_value()
        self._init_portfolio_values_if_necessary(force_recompute_origin_portfolio)

    def get_origin_portfolio_current_value(self, refresh_values=False):
        """
        Calculates and return the origin portfolio actual value
        :param refresh_values: when True, force origin portfolio reevaluation
        :return: the origin portfolio current value
        """
        if refresh_values:
            self.current_crypto_currencies_values.update(
                self._evaluate_config_crypto_currencies_and_portfolio_values(self.origin_portfolio.portfolio)
            )
        return self._update_portfolio_current_value(
            self.origin_portfolio.portfolio, currencies_values=self.current_crypto_currencies_values
        )

    def _init_portfolio_values_if_necessary(self, force_recompute_origin_portfolio):
        """
        Init origin portfolio values if necessary
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        if self.portfolio_origin_value == constants.ZERO:
            # try to update portfolio origin value if it's not known yet
            self._init_origin_portfolio_and_currencies_value()
        if force_recompute_origin_portfolio:
            self._recompute_origin_portfolio_initial_value()

    def _init_origin_portfolio_and_currencies_value(self):
        """
        Initialize origin portfolio and the origin portfolio currencies values
        """
        self.origin_portfolio = self.origin_portfolio or copy.copy(self.portfolio_manager.portfolio)
        self.origin_crypto_currencies_values.update(
            self._evaluate_config_crypto_currencies_and_portfolio_values(
                self.origin_portfolio.portfolio,
                ignore_missing_currency_data=True
            )
        )
        self._recompute_origin_portfolio_initial_value()

    def _update_portfolio_current_value(self, portfolio, currencies_values=None, fill_currencies_values=False):
        """
        Update the portfolio with current prices
        :param portfolio: the portfolio to update
        :param currencies_values: the currencies values
        :param fill_currencies_values: the currencies values to calculate
        :return: the updated portfolio
        """
        values = currencies_values
        if values is None or fill_currencies_values:
            value_update = self._evaluate_config_crypto_currencies_and_portfolio_values(portfolio)
            self.current_crypto_currencies_values.update(value_update)
            if len(self.current_crypto_currencies_values) > len(self.origin_crypto_currencies_values):
                # add any missing value to origin_crypto_currencies_values (can happen with indirect valuations)
                self._fill_currencies_values(self.origin_crypto_currencies_values)
            if fill_currencies_values:
                self._fill_currencies_values(currencies_values)
            values = self.current_crypto_currencies_values
        return self._evaluate_portfolio_value(portfolio, values)

    def _fill_currencies_values(self, currencies_values):
        """
        Fill a currency values dict with new data
        :param currencies_values: currencies values dict to be filled
        """
        currencies_values.update({
            currency: value
            for currency, value in self.current_crypto_currencies_values.items()
            if currency not in currencies_values
        })

    def _update_portfolio_and_currencies_current_value(self):
        """
        Update the portfolio current value with the current portfolio instance
        """
        self.portfolio_current_value = self._update_portfolio_current_value(
            self.portfolio_manager.portfolio.portfolio)

    def _recompute_origin_portfolio_initial_value(self):
        """
        Compute origin portfolio initial value and update portfolio_origin_value
        """
        if self.portfolio_manager.historical_portfolio_value_manager is not None \
           and self.portfolio_manager.historical_portfolio_value_manager.has_historical_starting_portfolio_value(
            self.portfolio_manager.reference_market
        ):
            # get origin value from history when possible
            value = self.portfolio_manager.historical_portfolio_value_manager.\
                get_historical_starting_starting_portfolio_value(self.portfolio_manager.reference_market)
            if value is not None:
                self.portfolio_origin_value = value
                return
        self.portfolio_origin_value = self._update_portfolio_current_value(
            self.origin_portfolio.portfolio,
            currencies_values=self.origin_crypto_currencies_values,
            fill_currencies_values=True
        )

    def _evaluate_config_crypto_currencies_and_portfolio_values(self,
                                                                portfolio,
                                                                ignore_missing_currency_data=False):
        """
        Evaluate both config and portfolio currencies values
        :param portfolio: the current portfolio
        :param ignore_missing_currency_data: when True, ignore missing currencies values in calculation
        :return: the result of config and portfolio currencies values calculation
        """
        evaluated_pair_values = {}
        evaluated_currencies = set()
        missing_tickers = set()

        self._evaluate_config_currencies_values(evaluated_pair_values, evaluated_currencies, missing_tickers)
        self._evaluate_portfolio_currencies_values(portfolio, evaluated_pair_values, evaluated_currencies,
                                                   missing_tickers, ignore_missing_currency_data)
        return evaluated_pair_values

    def _evaluate_config_currencies_values(self, evaluated_pair_values, evaluated_currencies, missing_tickers):
        """
        Evaluate config currencies values
        :param evaluated_pair_values: the list of evaluated pairs
        :param evaluated_currencies: the list of evaluated currencies
        :param missing_tickers: the list of missing currencies
        """
        if self.portfolio_manager.exchange_manager.exchange_config.traded_symbols:
            currency, market = \
                self.portfolio_manager.exchange_manager.exchange_config.traded_symbols[0].base_and_quote()
            currency_to_evaluate = currency
            try:
                if currency not in evaluated_currencies:
                    evaluated_pair_values[currency] = self.value_converter.evaluate_value(currency, constants.ONE)
                    evaluated_currencies.add(currency)
                if market not in evaluated_currencies:
                    currency_to_evaluate = market
                    evaluated_pair_values[market] = self.value_converter.evaluate_value(market, constants.ONE)
                    evaluated_currencies.add(market)
            except errors.MissingPriceDataError:
                missing_tickers.add(currency_to_evaluate)

    def _evaluate_portfolio_currencies_values(self,
                                              portfolio,
                                              evaluated_pair_values,
                                              evaluated_currencies,
                                              missing_tickers,
                                              ignore_missing_currency_data):
        """
        Evaluate current portfolio currencies values
        :param portfolio: the current portfolio
        :param evaluated_pair_values: the list of evaluated pairs
        :param evaluated_currencies: the list of evaluated currencies
        :param missing_tickers: the list of missing currencies
        :param ignore_missing_currency_data: when True, ignore missing currencies values in calculation
        """
        for currency in portfolio:
            try:
                if currency not in evaluated_currencies and self._should_currency_be_considered(
                        currency, portfolio, ignore_missing_currency_data
                ):
                    evaluated_pair_values[currency] = self.value_converter.evaluate_value(currency, constants.ONE)
                    evaluated_currencies.add(currency)
            except errors.MissingPriceDataError:
                missing_tickers.add(currency)

    def _evaluate_portfolio_value(self, portfolio, currencies_values=None):
        """
        Perform evaluate_value with a portfolio configuration
        :param portfolio: the portfolio to explore
        :param currencies_values: currencies to evaluate
        :return: the calculated quantity value in reference (attribute) currency
        """
        return sum([
            self._get_currency_value(portfolio, currency, currencies_values)
            for currency in portfolio
            if currency not in self.value_converter.missing_currency_data_in_exchange
        ])

    def _get_currency_value(self, portfolio, currency, currencies_values=None, raise_error=False):
        """
        Return the currency value
        :param portfolio: the specified portfolio
        :param currency: the currency to evaluate
        :param currencies_values: currencies values dict
        :param raise_error: When True, forward exceptions
        :return: the currency value
        """
        if currency in portfolio and portfolio[currency].total != constants.ZERO:
            if currencies_values and currency in currencies_values:
                return currencies_values[currency] * portfolio[currency].total
            return self.value_converter.evaluate_value(currency, portfolio[currency].total, raise_error)
        return constants.ZERO

    def _should_currency_be_considered(self, currency, portfolio, ignore_missing_currency_data):
        """
        Return True if enough data is available to evaluate currency value
        :param currency: the currency to evaluate
        :param portfolio: the specified portfolio
        :param ignore_missing_currency_data: When True, ignore check of currency presence
        in missing_currency_data_in_exchange
        :return: True if enough data is available to evaluate currency value
        """
        return (
            currency not in self.value_converter.missing_currency_data_in_exchange or ignore_missing_currency_data
        ) and (
            portfolio[currency].total > constants.ZERO
            or currency in self.portfolio_manager.portfolio_profitability.valuated_currencies
        )

    def clear(self):
        self.value_converter.clear()
        self.value_converter = None
        self.portfolio_manager = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import decimal

import octobot_commons.symbols as commons_symbols
import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data as trading_personal_data
import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.script_keywords.dsl as dsl
import octobot_trading.modes.script_keywords.basic_keywords.account_balance as account_balance

from tests import event_loop
from tests.modes.script_keywords import null_context

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_amount_from_input_amount(null_context):
    null_context.exchange_manager = mock.Mock(
        exchange_personal_data=mock.Mock(
            portfolio_manager=mock.Mock(
                portfolio_value_holder=mock.Mock()
            )
        )
    )
    with pytest.raises(errors.InvalidArgumentError):
        await script_keywords.get_amount_from_input_amount(null_context, "-1")

    with pytest.raises(errors.InvalidArgumentError):
        await script_keywords.get_amount_from_input_amount(null_context, "1sdsqdq")

    with pytest.raises(NotImplementedError):
        await script_keywords.get_amount_from_input_amount(null_context,
                                                           f"1{script_keywords.QuantityType.POSITION_PERCENT.value}")

    with mock.patch.object(account_balance, "adapt_amount_to_holdings",
                           mock.AsyncMock(return_value=decimal.Decimal(1))) as adapt_amount_to_holdings_mock:
        for quantity_delta in (script_keywords.QuantityType.DELTA, script_keywords.QuantityType.DELTA_BASE):
            with mock.patch.object(dsl, "parse_quantity",
                                   mock.Mock(return_value=(quantity_delta, decimal.Decimal(2)))) \
                    as parse_quantity_mock:
                assert await script_keywords.get_amount_from_input_amount(null_context, "1", "buy") \
                       == decimal.Decimal(1)
                adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal(2), "buy",
                                                                      False, True, False, target_price=None,
                                                                      orders_to_be_ignored=None)
                parse_quantity_mock.assert_called_once_with("1")
                parse_quantity_mock.reset_mock()
                adapt_amount_to_holdings_mock.reset_mock()

                with pytest.raises(errors.MissingFunds):
                    # raises as 1 != 2
                    await script_keywords.get_amount_from_input_amount(
                        null_context, "1", "buy", allow_holdings_adaptation=False, orders_to_be_ignored=[]
                    )
                adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal(2), "buy",
                                                                      False, True, False, target_price=None,
                                                                      orders_to_be_ignored=[])
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()

            with mock.patch.object(dsl, "parse_quantity",
                                   mock.Mock(return_value=(quantity_delta, decimal.Decimal(1)))) \
                    as parse_quantity_mock:
                for allow_holdings_adaptation in (True, False):
                    # doesn't raise as 1 == 1
                    assert await script_keywords.get_amount_from_input_amount(
                        null_context, "1", "buy", allow_holdings_adaptation=allow_holdings_adaptation,
                        orders_to_be_ignored=["plip"]
                    ) == decimal.Decimal(1)
                    adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal(1), "buy",
                                                                          False, True, False, target_price=None,
                                                                          orders_to_be_ignored=["plip"])
                    parse_quantity_mock.assert_called_once_with("1")
                    parse_quantity_mock.reset_mock()
                    adapt_amount_to_holdings_mock.reset_mock()

            with mock.patch.object(dsl, "parse_quantity",
                                   mock.Mock(return_value=(quantity_delta, decimal.Decimal("0.5")))) \
                    as parse_quantity_mock:
                for allow_holdings_adaptation in (True, False):
                    # doesn't raise as 1 > 0.5
                    assert await script_keywords.get_amount_from_input_amount(
                        null_context, "1", "buy", allow_holdings_adaptation=allow_holdings_adaptation
                    ) == decimal.Decimal(1)
                    adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("0.5"), "buy",
                                                                          False, True, False, target_price=None,
                                                                          orders_to_be_ignored=None)
                    parse_quantity_mock.assert_called_once_with("1")
                    parse_quantity_mock.reset_mock()
                    adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.DELTA_QUOTE, decimal.Decimal(2)))) \
                as parse_quantity_mock:
            with mock.patch.object(trading_personal_data, "get_up_to_date_price",
                                   mock.AsyncMock(return_value=decimal.Decimal(100))) \
                    as get_up_to_date_price_mock:
                assert await script_keywords.get_amount_from_input_amount(null_context, "1", "buy") == decimal.Decimal(1)
                get_up_to_date_price_mock.assert_awaited_with(
                    null_context.exchange_manager,
                    symbol=null_context.symbol,
                    timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
                )
                # without target_price
                adapt_amount_to_holdings_mock.assert_called_once_with(
                    null_context, decimal.Decimal(2) / decimal.Decimal(100), "buy",
                    False, True, False, target_price=None, orders_to_be_ignored=None
                )
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()
                parse_quantity_mock.reset_mock()
                get_up_to_date_price_mock.reset_mock()
                # with target_price
                assert await script_keywords.get_amount_from_input_amount(
                    null_context, "1", "buy", target_price=decimal.Decimal(23), allow_holdings_adaptation=False
                ) == decimal.Decimal(1)
                get_up_to_date_price_mock.assert_not_called()
                adapt_amount_to_holdings_mock.assert_called_once_with(
                    null_context, decimal.Decimal(2) / decimal.Decimal(23), "buy",
                    False, True, False, target_price=decimal.Decimal(23), orders_to_be_ignored=None
                )
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()
                parse_quantity_mock.reset_mock()

                with pytest.raises(errors.MissingFunds):
                    # raises as 1 < 4
                    await script_keywords.get_amount_from_input_amount(
                        null_context, "1", "buy",  target_price=decimal.Decimal("0.5"), allow_holdings_adaptation=False
                    )
                adapt_amount_to_holdings_mock.assert_called_once_with(
                    null_context, decimal.Decimal(4), "buy", False, True, False, target_price=decimal.Decimal("0.5"),
                    orders_to_be_ignored=None
                )
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.PERCENT, decimal.Decimal(75)))) \
                as parse_quantity_mock, \
                mock.patch.object(account_balance, "available_account_balance",
                                  mock.AsyncMock(return_value=decimal.Decimal(2))) \
                as available_account_balance_mock:
            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy", use_total_holding=True,
                                                                      reduce_only=False,
                                                                      is_stop_order=True,
                                                                      target_price=constants.ZERO) == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("1.5"), "buy",
                                                                  True, False, True, target_price=constants.ZERO,
                                                                  orders_to_be_ignored=None)
            parse_quantity_mock.assert_called_once_with("50")
            available_account_balance_mock.assert_called_once_with(null_context, "buy",
                                                                   use_total_holding=True, reduce_only=False)
            adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(
                                       script_keywords.QuantityType.AVAILABLE_PERCENT, decimal.Decimal(75)))) \
                as parse_quantity_mock, \
                mock.patch.object(account_balance, "available_account_balance",
                                  mock.AsyncMock(return_value=decimal.Decimal(2))) \
                as available_account_balance_mock:
            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("1.5"), "buy",
                                                                  False, True, False, target_price=None,
                                                                  orders_to_be_ignored=None)
            parse_quantity_mock.assert_called_once_with("50")
            available_account_balance_mock.assert_called_once_with(null_context, "buy",
                                                                   use_total_holding=False, reduce_only=True)
            adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(
                                       script_keywords.QuantityType.CURRENT_SYMBOL_ASSETS_PERCENT, decimal.Decimal(75)))) \
                as parse_quantity_mock, mock.patch.object(
            null_context.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder,
            "get_assets_holdings_value", mock.Mock(return_value=decimal.Decimal(2))
        ) as get_holdings_value_mock:
            null_context.symbol = None
            with pytest.raises(errors.InvalidArgumentError):
                # context.symbol is None
                assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(
                    1)
            parse_quantity_mock.reset_mock()
            null_context.symbol = "BTC/USDT"

            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("1.5"), "buy",
                                                                  False, True, False, target_price=None,
                                                                  orders_to_be_ignored=None)
            parse_quantity_mock.assert_called_once_with("50")
            get_holdings_value_mock.assert_called_once_with(("BTC", "USDT"), "BTC")
            adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(
                                       script_keywords.QuantityType.TRADED_SYMBOLS_ASSETS_PERCENT, decimal.Decimal(75)))) \
            as parse_quantity_mock, mock.patch.object(
                null_context.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder,
                "get_assets_holdings_value", mock.Mock(return_value=decimal.Decimal(10))
            ) as get_holdings_value_mock:
            null_context.symbol = None
            with pytest.raises(errors.InvalidArgumentError):
                # context.symbol is None
                assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(
                    1)
            parse_quantity_mock.reset_mock()
            null_context.symbol = "BTC/USDT"
            exchange_personal_data = null_context.exchange_manager.exchange_personal_data
            null_context.exchange_manager = mock.Mock(
                exchange_config=mock.Mock(
                    traded_symbols=[
                        commons_symbols.parse_symbol("BTC/USDT"),
                        commons_symbols.parse_symbol("ETH/BTC"),
                        commons_symbols.parse_symbol("SOL/USDT"),
                    ],
                ),
                exchange_personal_data=exchange_personal_data
            )

            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("7.5"), "buy",
                                                                  False, True, False, target_price=None,
                                                                  orders_to_be_ignored=None)
            parse_quantity_mock.assert_called_once_with("50")
            get_holdings_value_mock.assert_called_once_with({'BTC', 'ETH', 'SOL', 'USDT'}, "BTC")
            adapt_amount_to_holdings_mock.reset_mock()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data.orders.order as order_class
import octobot_trading.errors as errors


class UnsupportedOrder(order_class.Order):
    """
    UnsupportedOrder is used when an exchange order is fetched but is on purpose set to unsupported.
    Used to handle orders that are fetched but can't (yet) be handled in OctoBot.
    """
    async def update_order_status(self, force_refresh=False):
        if self.trader.simulate:
            # SHOULD NEVER HAPPEN
            raise errors.NotSupported(
                f"{self.get_name()} can't be updated and should not appear in simulation mode"
            )
        await self.default_exchange_update_order_status()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchanges.util import exchange_market_status_fixer
from octobot_trading.exchanges.util.exchange_market_status_fixer import (
    ExchangeMarketStatusFixer,
    is_ms_valid,
)
from octobot_trading.exchanges.util import symbol_details
from octobot_trading.exchanges.util.symbol_details import (
    SymbolDetails,
)
from octobot_trading.exchanges.util import exchange_util
from octobot_trading.exchanges.util.exchange_util import (
    get_rest_exchange_class,
    get_order_side,
    log_time_sync_error,
    get_partners_explanation_message,
    get_enabled_exchanges,
    get_local_exchange_manager,
    exchange_error_translator,
    is_compatible_account,
    get_historical_ohlcv,
    get_exchange_type,
    get_default_exchange_type,
    get_supported_exchange_types,
    update_raw_order_from_raw_trade,
    is_missing_trading_fees,
    apply_trades_fees,
    get_common_traded_quote,
    get_associated_symbol,
    get_exchange_class_from_name,
    get_auto_filled_exchange_names,
    get_exchange_details,
    is_error_on_this_type,
)
from octobot_trading.exchanges.util import websockets_util
from octobot_trading.exchanges.util.websockets_util import (
    force_disable_web_socket,
    check_web_socket_config,
    search_websocket_class,
    supports_websocket,
)

__all__ = [
    "ExchangeMarketStatusFixer",
    "is_ms_valid",
    "SymbolDetails",
    "get_rest_exchange_class",
    "get_order_side",
    "log_time_sync_error",
    "get_partners_explanation_message",
    "get_enabled_exchanges",
    "get_local_exchange_manager",
    "exchange_error_translator",
    "is_compatible_account",
    "get_historical_ohlcv",
    "get_exchange_type",
    "get_default_exchange_type",
    "get_supported_exchange_types",
    "update_raw_order_from_raw_trade",
    "is_missing_trading_fees",
    "apply_trades_fees",
    "get_common_traded_quote",
    "get_associated_symbol",
    "get_exchange_class_from_name",
    "get_auto_filled_exchange_names",
    "get_exchange_details",
    "is_error_on_this_type",
    "force_disable_web_socket",
    "check_web_socket_config",
    "search_websocket_class",
    "supports_websocket",
]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import pytest_asyncio
import numpy as np

from octobot_commons.enums import PriceIndexes, TimeFrames
from octobot_trading.api.symbol_data import get_symbol_candles_manager
from octobot_trading.exchange_data.ohlcv.candles_adapter import get_symbol_close_candles, get_symbol_open_candles, \
    get_symbol_low_candles, get_symbol_high_candles, get_symbol_time_candles, get_symbol_volume_candles, \
    get_candle_as_list
from octobot_trading.exchange_data.ohlcv.candles_manager import CandlesManager
from octobot_trading.exchange_data.kline.kline_manager import KlineManager
from octobot_trading.exchange_data.exchange_symbol_data import ExchangeSymbolData
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from tests import event_loop


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def time_frame():
    return "1h"


@pytest_asyncio.fixture
async def symbol_data(time_frame):
    symbol_candles = CandlesManager()
    await symbol_candles.initialize()
    symbol_candles.replace_all_candles(_get_candles())
    symbol_kline = KlineManager()
    await symbol_kline.initialize()
    symbol_kline.kline_update(_get_candle(11))
    manager = ExchangeManager({}, "binanceus")
    symbol_data = ExchangeSymbolData(manager, "BTC/USDT")
    tf = TimeFrames(time_frame)
    symbol_data.symbol_candles[tf] = symbol_candles
    symbol_data.symbol_klines[tf] = symbol_kline
    return symbol_data


async def test_get_symbol_close_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_close_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_CLOSE.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_close_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_CLOSE.value),
                                   dtype=float))


async def test_get_symbol_open_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_open_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_OPEN.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_open_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_OPEN.value),
                                   dtype=float))


async def test_get_symbol_low_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_low_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_LOW.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_low_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_LOW.value),
                                   dtype=float))


async def test_get_symbol_high_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_high_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_HIGH.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_high_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_HIGH.value),
                                   dtype=float))


async def test_get_symbol_time_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_time_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_TIME.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_time_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_TIME.value),
                                   dtype=float))


async def test_get_symbol_volume_candles(symbol_data, time_frame):
    # default selector
    assert np.array_equal(get_symbol_volume_candles(symbol_data, time_frame, 10, False),
                          np.array(_get_candles_extract(PriceIndexes.IND_PRICE_VOL.value), dtype=float))

    # selector with in construction candle
    assert np.array_equal(get_symbol_volume_candles(symbol_data, time_frame, 10, True),
                          np.array(_get_candles_extract_with_extra_candle(PriceIndexes.IND_PRICE_VOL.value),
                                   dtype=float))


async def test_get_candle_as_list(symbol_data, time_frame):
    row_candles = _get_candles()
    candles = get_symbol_candles_manager(symbol_data, time_frame).get_symbol_prices()
    candle_as_list = get_candle_as_list(candles, 0)
    assert candle_as_list == row_candles[0]

    candles_limit_1 = get_symbol_candles_manager(symbol_data, time_frame).get_symbol_prices(limit=1)
    candles_list_1 = get_candle_as_list(candles_limit_1, 0)
    assert candles_list_1 == row_candles[-1]

    candles_limit_2 = get_symbol_candles_manager(symbol_data, time_frame).get_symbol_prices(limit=2)
    candles_list_2 = get_candle_as_list(candles_limit_2, 0)
    assert candles_list_2 == row_candles[-2]

    candles_limit_2 = get_symbol_candles_manager(symbol_data, time_frame).get_symbol_prices(limit=2)
    candles_list_2 = get_candle_as_list(candles_limit_2, 1)
    assert candles_list_2 == row_candles[-1]


def _get_candles_extract(index, count=10):
    return [elem[index] for elem in _get_candles(count)]


def _get_candles_extract_with_extra_candle(index, count=10):
    return [elem[index] for elem in _get_candles(count=count+1)][1:]


def _get_candles(count=10):
    return [_get_candle(seed) for seed in range(1, count + 1)]


def _get_candle(seed):
    return [seed * (i + 1) * 1 / 3 for i, _ in enumerate(PriceIndexes)]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock


def order_mock(**kwargs):
    order = mock.Mock(**kwargs)
    order.is_open = mock.Mock(return_value=True)
    order.is_cancelling = mock.Mock(return_value=False)
    order.trader = mock.Mock()
    order.trader.cancel_order = mock.AsyncMock()
    order.trader.edit_order = mock.AsyncMock()
    order.trader.exchange_manager = mock.Mock(trading_modes=[])
    order.trader.exchange_manager.trader = order.trader
    return order



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import decimal

import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.script_keywords.dsl as dsl
import octobot_trading.modes.script_keywords.basic_keywords.position as position
import octobot_trading.personal_data as personal_data
import octobot_trading.errors as errors
import octobot_trading.constants as constants
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc

from tests import event_loop
from tests.modes.script_keywords import null_context

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_price_with_offset(null_context):
    null_context.symbol = "blop/plop"
    null_context.exchange_manager = mock.Mock(
        exchange=mock.Mock(
            get_market_status=mock.Mock(return_value={
                Ecmsc.PRECISION.value: {
                    Ecmsc.PRECISION_PRICE.value: 2
                }
            })
        )
    )

    with pytest.raises(errors.InvalidArgumentError):
        await script_keywords.get_price_with_offset(null_context, "1sdsqdq")

    with mock.patch.object(personal_data, "get_up_to_date_price", mock.AsyncMock(return_value=200)) \
            as current_price_mock:
        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.DELTA, decimal.Decimal(10)))) \
                as parse_quantity_mock:
            assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(210)
            current_price_mock.assert_called_once_with(
                null_context.exchange_manager, null_context.symbol, timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
            )
            parse_quantity_mock.assert_called_once_with(10)
            null_context.exchange_manager.exchange.get_market_status.assert_called_once_with(
                "blop/plop", with_fixer=False
            )
            current_price_mock.reset_mock()
            parse_quantity_mock.reset_mock()

            assert await script_keywords.get_price_with_offset(null_context, 10, use_delta_type_as_flat_value=True) \
                   == decimal.Decimal(10)
            current_price_mock.assert_not_called()
            parse_quantity_mock.assert_called_once_with(10)
            current_price_mock.reset_mock()
            parse_quantity_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.DELTA_EXPLICIT, decimal.Decimal(10)))) \
                as parse_quantity_mock:
            assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(210)
            current_price_mock.assert_called_once_with(
                null_context.exchange_manager, null_context.symbol, timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
            )
            parse_quantity_mock.assert_called_once_with(10)
            current_price_mock.reset_mock()
            parse_quantity_mock.reset_mock()

            # DELTA_EXPLICIT ignores use_delta_type_as_flat_value
            assert await script_keywords.get_price_with_offset(null_context, 10, use_delta_type_as_flat_value=True) \
                   == decimal.Decimal(210)
            current_price_mock.assert_called_once_with(
                null_context.exchange_manager, null_context.symbol, timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
            )
            parse_quantity_mock.assert_called_once_with(10)
            current_price_mock.reset_mock()
            parse_quantity_mock.reset_mock()

        with mock.patch.object(
            dsl, "parse_quantity",
            mock.Mock(return_value=(script_keywords.QuantityType.DELTA, decimal.Decimal("10.333333")))
        ) as parse_quantity_mock:
            assert await script_keywords.get_price_with_offset(
                null_context, 10.333333, use_delta_type_as_flat_value=True
            ) == decimal.Decimal("10.33")   # rounded according to precision
            current_price_mock.assert_not_called()
            parse_quantity_mock.assert_called_once_with(10.333333)
            current_price_mock.reset_mock()
            parse_quantity_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.PERCENT, decimal.Decimal(-99)))):
            assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(2)

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.PERCENT, decimal.Decimal(1000)))):
            assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(2200)

    with mock.patch.object(position, "average_open_pos_entry", mock.AsyncMock(return_value=500)) \
            as average_open_pos_entry_mock:
        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.ENTRY_PERCENT, decimal.Decimal(-50)))):
            assert await script_keywords.get_price_with_offset(null_context, 10, side="sell") == decimal.Decimal(250)
            average_open_pos_entry_mock.assert_called_once_with(null_context, "sell")

    with mock.patch.object(position, "average_open_pos_entry", mock.AsyncMock(return_value=500)) \
            as average_open_pos_entry_mock:
        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.ENTRY, decimal.Decimal(50)))):
            assert await script_keywords.get_price_with_offset(null_context, 10, side="sell") == decimal.Decimal(550)
            average_open_pos_entry_mock.assert_called_once()

    with mock.patch.object(dsl, "parse_quantity",
                           mock.Mock(return_value=(script_keywords.QuantityType.FLAT, decimal.Decimal(50)))):
        assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(50)

    with mock.patch.object(dsl, "parse_quantity",
                           mock.Mock(return_value=(script_keywords.QuantityType.DELTA_QUOTE, decimal.Decimal(50)))):
        assert await script_keywords.get_price_with_offset(null_context, 10) == decimal.Decimal(50)

    with mock.patch.object(dsl, "parse_quantity",
                           mock.Mock(return_value=(script_keywords.QuantityType.UNKNOWN, decimal.Decimal(-50)))):
        with pytest.raises(errors.InvalidArgumentError):
            await script_keywords.get_price_with_offset(null_context, 10)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import copy
import decimal
import time

import octobot_commons.channels_name as channels_name
import octobot_commons.enums as commons_enums
import octobot_commons.databases as commons_databases
import octobot_commons.logging as commons_logging
import octobot_commons.authentication as authentication

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.storage.util as storage_util


class OrdersStorage(abstract_storage.AbstractStorage):
    LIVE_CHANNEL = channels_name.OctoBotTradingChannelsName.ORDERS_CHANNEL.value
    HISTORY_TABLE = commons_enums.DBTables.ORDERS.value
    HISTORICAL_OPEN_ORDERS_TABLE = commons_enums.DBTables.HISTORICAL_ORDERS_UPDATES.value
    ENABLE_HISTORICAL_ORDER_UPDATES_STORAGE = constants.ENABLE_HISTORICAL_ORDERS_UPDATES_STORAGE

    def __init__(self, exchange_manager, use_live_consumer_in_backtesting=None, is_historical=None):
        super().__init__(exchange_manager, plot_settings=None,
                         use_live_consumer_in_backtesting=use_live_consumer_in_backtesting, is_historical=is_historical)
        self.startup_orders = {}

    def should_store_data(self):
        return not self.exchange_manager.is_trader_simulated \
            and not self.exchange_manager.is_backtesting

    async def on_start(self):
        await self._load_startup_orders()

    async def _live_callback(
        self,
        exchange: str,
        exchange_id: str,
        cryptocurrency: str,
        symbol: str,
        order: dict,
        update_type: str,
        is_from_bot: bool,
    ):
        await self.trigger_debounced_update_auth_data(False)
        # only store the current snapshot of open orders when order updates are received
        if self.should_store_data():
            await self._update_history()
            if self.ENABLE_HISTORICAL_ORDER_UPDATES_STORAGE:
                await self._add_historical_open_orders(order, update_type)
            await self.trigger_debounced_flush()

    @abstract_storage.AbstractStorage.hard_reset_and_retry_if_necessary
    async def _update_history(self):
        await self._get_db().replace_all(
            self.HISTORY_TABLE,
            [
                _format_order(order, self.exchange_manager)
                for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders()
            ],
            cache=False,
        )

    async def _add_historical_open_orders(self, order_dict: dict, update_type: str):
        update_time = time.time()
        await self._get_db().log(
            self.HISTORICAL_OPEN_ORDERS_TABLE,
            _format_order_update(self.exchange_manager, order_dict, update_type, update_time),
            cache=False,
        )

    async def _update_auth_data(self, _):
        authenticator = authentication.Authenticator.instance()
        snapshot = [
            _format_order(order, self.exchange_manager)
            for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders()
        ]
        if authenticator.is_initialized():
            # also update when history is empty to reset trade history
            await authenticator.update_orders(snapshot, self.exchange_manager.exchange_name)

    async def _store_history(self):
        await self._update_history()
        await self._get_db().flush()

    def _get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_orders_db(
            self.exchange_manager.bot_id,
            storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager),
            self.exchange_manager.exchange_name,
        )

    async def get_historical_orders_updates(self):
        return copy.deepcopy(await self._get_db().all(self.HISTORICAL_OPEN_ORDERS_TABLE))

    async def get_startup_order_details(self, order_exchange__id):
        return self.startup_orders.get(order_exchange__id, None)

    async def _load_startup_orders(self):
        if self.should_store_data():
            self.startup_orders = {
                _get_startup_order_key(order): from_order_document(order)
                for order in copy.deepcopy(await self._get_db().all(self.HISTORY_TABLE))
                if order    # skip empty order details (error when serializing)
            }
        else:
            self.startup_orders = {}

    def get_startup_self_managed_orders_details_from_group(self, group_id):
        return [
            order
            for order in self.get_all_self_managed_startup_orders()
            if order.get(enums.StoredOrdersAttr.GROUP.value, {}).get(enums.StoredOrdersAttr.GROUP_ID.value, None)
            == group_id
        ]

    def get_all_self_managed_startup_orders(self):
        return [
            order
            for order in self.startup_orders.values()
            if order.get(constants.STORAGE_ORIGIN_VALUE, {})
            .get(enums.ExchangeConstantsOrderColumns.SELF_MANAGED.value, False)
        ]

    @classmethod
    async def clear_database_history(cls, database, flush=True):
        await super().clear_database_history(database, flush=False)
        if cls.ENABLE_HISTORICAL_ORDER_UPDATES_STORAGE:
            await database.delete(cls.HISTORICAL_OPEN_ORDERS_TABLE, None)
        if flush:
            await database.flush()


def _get_group_dict(order):
    if not order.order_group:
        return {}
    try:
        return {
            enums.StoredOrdersAttr.GROUP_ID.value: order.order_group.name,
            enums.StoredOrdersAttr.GROUP_TYPE.value: order.order_group.__class__.__name__,
        }
    except KeyError:
        return {}


def _get_chained_orders(order, exchange_manager):
    if not order.chained_orders:
        return []
    return [
        _format_order(chained_order, exchange_manager)
        for chained_order in order.chained_orders
    ]


def _format_order(order, exchange_manager):
    try:
        formatted = {
            constants.STORAGE_ORIGIN_VALUE: OrdersStorage.sanitize_for_storage(order.to_dict()),
        }
        for key, val in (
            (enums.StoredOrdersAttr.EXCHANGE_CREATION_PARAMS, OrdersStorage.sanitize_for_storage(order.exchange_creation_params)),
            (enums.StoredOrdersAttr.TRADER_CREATION_KWARGS, OrdersStorage.sanitize_for_storage(order.trader_creation_kwargs)),
            (enums.StoredOrdersAttr.HAS_BEEN_BUNDLED, order.has_been_bundled),
            (enums.StoredOrdersAttr.ENTRIES, order.associated_entry_ids),
            (enums.StoredOrdersAttr.GROUP, _get_group_dict(order)),
            (enums.StoredOrdersAttr.CHAINED_ORDERS, _get_chained_orders(order, exchange_manager)),
            (enums.StoredOrdersAttr.UPDATE_WITH_TRIGGERING_ORDER_FEES, order.update_with_triggering_order_fees),
        ):
            # do not include default values
            if val:
                formatted[key.value] = val
        return formatted
    except Exception as err:
        commons_logging.get_logger(OrdersStorage.__name__).exception(err, True, f"Error when formatting order: {err}")
    return {}


def _format_order_update(exchange_manager, order_dict, update_type, update_time):
    order_id = order_dict[enums.ExchangeConstantsOrderColumns.ID.value]
    status = order_dict[enums.ExchangeConstantsOrderColumns.STATUS.value]
    order_update = {
        enums.StoredOrdersAttr.ORDER_ID.value: order_id,
        enums.StoredOrdersAttr.ORDER_STATUS.value: status,
        enums.StoredOrdersAttr.UPDATE_TIME.value: update_time,
        enums.StoredOrdersAttr.UPDATE_TYPE.value: update_type,
    }
    details = None
    try:
        details = _format_order(
            exchange_manager.exchange_personal_data.orders_manager.get_order(order_id),
            exchange_manager
        )
    except KeyError:
        if status == enums.OrderStatus.OPEN.value:
            # ensure order details are present in open orders
            details = {
                constants.STORAGE_ORIGIN_VALUE: OrdersStorage.sanitize_for_storage(order_dict),
            }
    order_update[enums.StoredOrdersAttr.ORDER_DETAILS.value] = details
    return order_update


def from_order_document(order_document):
    order_dict = dict(order_document)
    try:
        restore_order_storage_origin_value(order_dict[constants.STORAGE_ORIGIN_VALUE])
        for chained_order in order_dict.get(enums.StoredOrdersAttr.CHAINED_ORDERS.value, []):
            from_order_document(chained_order)
    except Exception as err:
        commons_logging.get_logger(OrdersStorage.__name__).exception(
            err, True, f"Error when reading: {err} order: {order_document}"
        )
    return order_dict


def restore_order_storage_origin_value(origin_val):
    origin_val[enums.ExchangeConstantsOrderColumns.AMOUNT.value] = \
        decimal.Decimal(str(origin_val[enums.ExchangeConstantsOrderColumns.AMOUNT.value]))
    origin_val[enums.ExchangeConstantsOrderColumns.COST.value] = \
        decimal.Decimal(str(origin_val[enums.ExchangeConstantsOrderColumns.COST.value]))
    origin_val[enums.ExchangeConstantsOrderColumns.FILLED.value] = \
        decimal.Decimal(str(origin_val[enums.ExchangeConstantsOrderColumns.FILLED.value]))
    origin_val[enums.ExchangeConstantsOrderColumns.PRICE.value] = \
        decimal.Decimal(str(origin_val[enums.ExchangeConstantsOrderColumns.PRICE.value]))
    if origin_val[enums.ExchangeConstantsOrderColumns.FEE.value] and \
            enums.FeePropertyColumns.COST.value in origin_val[enums.ExchangeConstantsOrderColumns.FEE.value]:
        origin_val[enums.ExchangeConstantsOrderColumns.FEE.value][enums.FeePropertyColumns.COST.value] = \
            decimal.Decimal(str(
                origin_val[enums.ExchangeConstantsOrderColumns.FEE.value][enums.FeePropertyColumns.COST.value]
            ))
    return origin_val


def _get_startup_order_key(order_dict):
    # use exchange id if available, fallback to order_id (for self managed orders)
    return order_dict[constants.STORAGE_ORIGIN_VALUE][enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value] or \
        order_dict[constants.STORAGE_ORIGIN_VALUE][enums.ExchangeConstantsOrderColumns.ID.value]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
# from distutils.extension import Extension
from setuptools import find_packages
from setuptools import setup

from octobot_trading import PROJECT_NAME, VERSION

PACKAGES = find_packages(exclude=["tests"])

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()

REQUIRED = open('requirements.txt').readlines()
REQUIRES_PYTHON = '>=3.8'

setup(
    name=PROJECT_NAME,
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Trading',
    license='LGPL-3.0',
    author='Drakkar-Software',
    author_email='contact@drakkar.software',
    description='OctoBot project trading package',
    packages=PACKAGES,
    include_package_data=True,
    long_description=DESCRIPTION,
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    data_files=[],
    install_requires=REQUIRED,
    python_requires=REQUIRES_PYTHON,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX'
    ],
)
